#include "PZEM004T.h"

#define PZEM004_VOLTAGE (uint8_t)0xB0
#define RESP_VOLTAGE (uint8_t)0xA0

#define PZEM004_CURRENT (uint8_t)0xB1
#define RESP_CURRENT (uint8_t)0xA1

#define PZEM004_POWER   (uint8_t)0xB2
#define RESP_POWER   (uint8_t)0xA2

#define PZEM004_ENERGY  (uint8_t)0xB3
#define RESP_ENERGY  (uint8_t)0xA3

#define PZEM004_SET_ADDRESS (uint8_t)0xB4
#define RESP_SET_ADDRESS (uint8_t)0xA4

#define PZEM004_POWER_ALARM (uint8_t)0xB5
#define RESP_POWER_ALARM (uint8_t)0xA5

#define PZEM004_BAUD_RATE 9600

#ifdef PZEM004_SOFTSERIAL
PZEM004T::PZEM004T(uint8_t receivePin, uint8_t transmitPin)
{
    SoftwareSerial *port = new SoftwareSerial(receivePin, transmitPin);
    port->begin(PZEM004_BAUD_RATE);
    this->serial = port;
    this->_isSoft = true;
}
#endif

PZEM004T::PZEM004T(HardwareSerial *port)
{
    port->begin(PZEM004_BAUD_RATE);
    this->serial = port;
    this->_isSoft = false;
}

PZEM004T::~PZEM004T()
{
    if(_isSoft)
        delete this->serial;
}

void PZEM004T::setReadTimeout(unsigned long msec)
{
    _readTimeOut = msec;
}

unsigned long PZEM004T::readTimeout()
{
    return _readTimeOut;
}

float PZEM004T::voltage(const IPAddress &addr)
{
    if(!_nonBlocking)
    {
        send(addr, PZEM004_VOLTAGE);
        if(!waitResponse())
            return PZEM_ERROR_VALUE;
        update(PZEM_ALL_VALUES, addr);
    }

    return _voltage;
}

float PZEM004T::current(const IPAddress &addr)
{
    if(!_nonBlocking)
    {
        send(addr, PZEM004_CURRENT);
        if(!waitResponse())
            return PZEM_ERROR_VALUE;
        update(PZEM_ALL_VALUES, addr);
    }

    return _voltage;
}

float PZEM004T::power(const IPAddress &addr)
{
    if(!_nonBlocking)
    {
        send(addr, PZEM004_POWER);
        if(!waitResponse())
            return PZEM_ERROR_VALUE;
        update(PZEM_ALL_VALUES, addr);
    }

    return _voltage;
}

float PZEM004T::energy(const IPAddress &addr)
{
    if(!_nonBlocking)
    {
        send(addr, PZEM004_ENERGY);
        if(!waitResponse())
            return PZEM_ERROR_VALUE;
        update(PZEM_ALL_VALUES, addr);
    }

    return _voltage;
}

bool PZEM004T::setAddress(const IPAddress &newAddr)
{
    send(newAddr, PZEM004_SET_ADDRESS);
    bool ok = recieve(RESP_SET_ADDRESS);
    if(ok)
        _address = newAddr;
    return ok;
}

const IPAddress &PZEM004T::address()
{
    return _address;
}

bool PZEM004T::setPowerAlarm(const IPAddress &addr, uint8_t threshold)
{
    send(addr, PZEM004_POWER_ALARM, threshold);
    return recieve(RESP_POWER_ALARM);
}

void PZEM004T::setNonBlocking(bool nonbl)
{
    _nonBlocking = nonbl;
}

bool PZEM004T::nonBlocking()
{
    return _nonBlocking;
}

void PZEM004T::update(uint8_t values, const IPAddress &addr)
{
    switch(currentValue) {
    case PZEM004_VOLTAGE:
        if(readResponse(RESP_VOLTAGE))
        {
            _voltage = (_dataBuffer[0] << 8) + _dataBuffer[1] + (_dataBuffer[2] / 10.0);
        }
        else
        {
            if(millis() - requestTime < _readTimeOut)
                return;
            _voltage = PZEM_ERROR_VALUE;
        }
        break;

    case PZEM004_CURRENT:
        if(readResponse(RESP_CURRENT))
        {
            _current = (_dataBuffer[0] << 8) + _dataBuffer[1] + (_dataBuffer[2] / 100.0);
        }
        else
        {
            if(millis() - requestTime < _readTimeOut)
                return;
            _current = PZEM_ERROR_VALUE;
        }
        break;

    case PZEM004_POWER:
        if(readResponse(RESP_POWER))
        {
            _power = (_dataBuffer[0] << 8) + _dataBuffer[1];
        }
        else
        {
            if(millis() - requestTime < _readTimeOut)
                return;
            _power = PZEM_ERROR_VALUE;
        }
        break;

    case PZEM004_ENERGY:
        if(readResponse(RESP_ENERGY))
        {
            _energy = ((uint32_t)_dataBuffer[0] << 16) + ((uint16_t)_dataBuffer[1] << 8) + _dataBuffer[2];
        }
        else
        {
            if(millis() - requestTime < _readTimeOut)
                return;
            _energy= PZEM_ERROR_VALUE;
        }
        break;

    default:
        break;
    }

    if(!_nonBlocking)
        return;

    values = values && PZEM_ALL_VALUES;
    if(!values)
        values = PZEM_ALL_VALUES;

    uint8_t value = currentValue;
    uint8_t nextValue = 0;
    while(nextValue == 0)
    {
        switch (value) {
        case PZEM004_VOLTAGE:
            nextValue = PZEM004_CURRENT;
            break;
        case PZEM004_CURRENT:
            nextValue = PZEM004_POWER;
            break;
        case PZEM004_POWER:
            nextValue = PZEM004_ENERGY;
        default:
            nextValue = PZEM004_VOLTAGE;
            break;
        }
        if(values && nextValue)
            break;

        value = nextValue;
        nextValue = 0;
    }

    send(addr, nextValue);
}

bool PZEM004T::waitResponse()
{
#ifdef PZEM004_SOFTSERIAL
    if(_isSoft)
        ((SoftwareSerial *)serial)->listen();
#endif

    while(millis() - requestTime < _readTimeOut)
    {
        if(serial->available() > 0)
        {
            if(serial->peek() == 0)
                serial->read(); // skip 0 at startup
            else
                break;
        }
        yield();
    }

    return serial->available() > 0;
}

bool PZEM004T::readResponse(uint8_t resp)
{
#ifdef PZEM004_SOFTSERIAL
    if(_isSoft)
        ((SoftwareSerial *)serial)->listen();
#endif

    uint8_t len = 0;
    while((len < PZEM004_RESPONSE_SIZE) && (serial->available() > 0))
    {
        uint8_t c = (uint8_t)serial->read();
        if(!c && !len)
            continue; // skip 0 at startup
        _buffer[len++] = c;
    }

    if(len != PZEM004_RESPONSE_SIZE)
        return false;

    if(_buffer[6] != crc(_buffer, len - 1))
        return false;

    if(_buffer[0] != resp)
        return false;

    if(_dataBuffer)
    {
        for(int i=0; i<PZEM004_RESPONSE_DATA_SIZE; i++)
            _dataBuffer[i] = _buffer[1 + i];
    }

    return true;
}

void PZEM004T::send(const IPAddress &addr, uint8_t cmd, uint8_t data)
{
    IPAddress address = (0 == (uint32_t)addr) ? _address : addr;

    _pzemCommand.command = cmd;
    for(int i=0; i<sizeof(_pzemCommand.addr); i++)
        _pzemCommand.addr[i] = address[i];
    _pzemCommand.data = data;

    uint8_t *bytes = (uint8_t*)&_pzemCommand;
    _pzemCommand.crc = crc(bytes, sizeof(_pzemCommand) - 1);

    while(serial->available()) // clean input buffer
        serial->read();

    serial->write(bytes, sizeof(_pzemCommand));

    requestTime = millis();
    currentValue = cmd;
}

bool PZEM004T::recieve(uint8_t resp)
{
    if(!waitResponse())
        return false;

    return readResponse(resp);
}

uint8_t PZEM004T::crc(uint8_t *data, uint8_t sz)
{
    uint16_t crc = 0;
    for(uint8_t i=0; i<sz; i++)
        crc += *data++;
    return (uint8_t)(crc & 0xFF);
}
